import os
import warnings
from typing import AnyStr, Tuple, List, Dict

import openai
import requests
from PIL import Image

from jam.interface.base import BaseInterface, JIOutput, JIError
from jam.util.generate import generate_id

warnings.filterwarnings('ignore')


class OpenAIBase(BaseInterface):

    def __init__(self, token: str = None, model: str = None):
        super(OpenAIBase, self).__init__()
        self._token = token
        if self._token is None:
            self._token = os.getenv('OPENAI_KEY', '')

        if self._token == '':
            raise JIError(f"Unfulfilled credentials for {self.__class__.__name__} in parameters or environment.")

        openai.api_key = self._token
        self.open_ai = openai

        self.model = model


class OpenAIChat(OpenAIBase):
    DEFAULT_MODEL = 'gpt-3.5-turbo-0613'

    ROLE_SYSTEM = 'system'
    ROLE_USER = 'user'
    ROLE_ASSISTANT = 'assistant'
    ROLE_FUNCTION = 'function'

    def __init__(self,
                 token: str = None,
                 model: str = DEFAULT_MODEL,
                 functions: List[Dict] = None):
        super(OpenAIChat, self).__init__(token, model)
        self.functions = functions

    def call(self, messages: List[Dict] = None, functional: bool = True) -> JIOutput:
        x_message = self.form_message(messages)
        if functional:
            response = self.open_ai.ChatCompletion.create(
                model=self.model,
                messages=x_message,
                functions=self.functions,
                function_call='auto'
            )
            response_message = response['choices'][0]['message']
            return JIOutput(output=response_message)

        response = self.open_ai.ChatCompletion.create(
            model=self.model,
            messages=x_message,
        )
        response_message = response['choices'][0]['message']
        return JIOutput(output=response_message)

    @staticmethod
    def form_message(messages: List[Dict]) -> List[Dict]:
        msg_objs = []
        for message in messages:
            msg_obj = {
                'role': message.get('role', 'user'),
                'content': message.get('content', '')
            }
            if message.get('name'):
                msg_obj['name'] = message.get('name')

            msg_objs.append(msg_obj)
        return msg_objs


class OpenAIImageGen(OpenAIBase):

    def __init__(self,
                 token: str = None,
                 model: str = None,
                 samples: int = 1,
                 dimension: Tuple[int, int] = (512, 512)):
        super(OpenAIImageGen, self).__init__(token, model)

        self.samples = samples
        self.width, self.height = dimension

    def call(self, x: AnyStr = None) -> JIOutput:
        try:
            filenames = []
            response = openai.Image.create(
                prompt=x,
                n=self.samples,
                size=f'{self.width}x{self.height}'
            )

            for resp in response.get('data', []):
                filename = f'DALLE_{generate_id(8)}.png'
                img_url = resp.get('url', '')
                if img_url == '':
                    continue

                img = Image.open(requests.get(img_url, stream=True).raw)
                img.save(filename)
                filenames.append(filename)
            return JIOutput(
                out_type=JIOutput.JIO_TEXT,
                output=' '.join(filenames)
            )
        except openai.error.OpenAIError as oai_err:
            return JIOutput(
                out_type=JIOutput.JIO_TEXT,
                output=oai_err.user_message,
                success=False
            )



import os
import io
import warnings
from typing import AnyStr, Tuple

from jam.interface.base import BaseInterface, JIOutput, JIError

from PIL import Image
from stability_sdk import client
import stability_sdk.interfaces.gooseai.generation.generation_pb2 as generation


class StabilityAIBase(BaseInterface):

    def __init__(self,
                 engine: str,
                 api_key: str = None,
                 host: str = 'grpc.stability.ai:443',
                 verbose: bool = True,
                 step: int = 30,
                 cfg_scale: float = 8.0,
                 dimension: Tuple[int, int] = (512, 512),
                 samples: int = 1,
                 sampler: int = generation.SAMPLER_K_DPMPP_2M
                 ):
        super(StabilityAIBase, self).__init__()
        self.engine = engine
        self._api_key = api_key
        if self._api_key is None:
            self._api_key = os.getenv('STABILITY_KEY', '')

        self._host = host
        if self._host is None:
            self._host = os.getenv('STABILITY_HOST', '')

        if self._api_key == '':
            raise JIError(f"Unfulfilled credentials for {self.__class__.__name__} in parameters or environment.")

        self.verbose = verbose
        self.stability_api = client.StabilityInference(
            key=self._api_key,
            verbose=self.verbose,
            engine=self.engine,
        )

        # Generation Config
        self.step = step
        self.cfg_scale = cfg_scale
        self.width, self.height = dimension
        self.samples = samples
        self.sampler = sampler


class StabilityText2Image(StabilityAIBase):

    def __init__(self,
                 engine: str = 'stable-diffusion-v1',
                 api_key: str = None,
                 host: str = 'grpc.stability.ai:443',
                 verbose: bool = True):
        super(StabilityText2Image, self).__init__(engine, api_key, host, verbose)

    def call(self, x: AnyStr = None) -> JIOutput:
        filenames = []
        response = self.stability_api.generate(
            prompt=x,
            steps=self.step,
            cfg_scale=self.cfg_scale,
            width=self.width,
            height=self.height,
            samples=self.samples,
            sampler=self.sampler
        )

        for resp in response:
            for artifact in resp.artifacts:
                if artifact.finish_reason == generation.FILTER:
                    warnings.warn(
                        "Your request activated the API's safety filters and could not be processed."
                        "Please modify the prompt and try again.")
                if artifact.type == generation.ARTIFACT_IMAGE:
                    filename = 'Stability_' + str(artifact.seed) + ".png"
                    img = Image.open(io.BytesIO(artifact.binary))
                    img.save(filename)
                    filenames.append(filename)
        return JIOutput(
            out_type=JIOutput.JIO_TEXT,
            output=' '.join(filenames)
        )



from typing import Union, Dict, AnyStr

from jam.instrument.base import BaseInstrument, BaseParameter, BaseProperty
from jam.interface.base import BaseInterface
from jam.interface.openai import OpenAIImageGen

_PromptPainterPromptProperty = BaseProperty(
    name='prompt', description='The text prompt to generate the image'
)
_PromptPainterParameters = BaseParameter(
    properties=[_PromptPainterPromptProperty],
    required=[_PromptPainterPromptProperty]
)


class PromptPainter(BaseInstrument):

    def __init__(self,
                 name: str = 'prompt_painter',
                 description: str = 'Generates image from text prompt',
                 parameters: Union[BaseParameter, Dict] = _PromptPainterParameters,
                 interface: BaseInterface = None):
        super().__init__(name, description, parameters, interface)
        if self.interface is None:
            self.interface = OpenAIImageGen()

    def activate(self, prompt: AnyStr = None):
        if prompt is None:
            return

        response = self.interface.call(prompt)
        return response



from jam.interface.openai import OpenAIChat
from jam.interface.openai import OpenAIImageGen
from jam.interface.stability_ai import StabilityText2Image
from jam.interface.writesonic import PhotoSonicText2Image



import unittest
from datetime import datetime
from typing import AnyStr, List

from jam.persistence.base import BasePersistence
from jam.persistence.base import PersistenceObject

from jam.util.generate import generate_id


class TestPersistenceObject(unittest.TestCase):

    def setUp(self) -> None:
        self.uid_1 = generate_id(16)
        self.author_1 = generate_id(8)
        self.role_1 = generate_id(8)
        self.content_1 = generate_id(8)
        self.mention_1 = self.author_1
        self.function_1 = generate_id(8)
        self.timestamp_1 = datetime.now()
        self.success_1 = True

        self.uid_2 = generate_id(16)
        self.author_2 = generate_id(8)
        self.role_2 = generate_id(8)
        self.content_2 = generate_id(8)
        self.mention_2 = self.author_2
        self.function_2 = generate_id(8)
        self.timestamp_2 = datetime.now()
        self.success_2 = False

        self.po_1 = PersistenceObject(
            uid=self.uid_1,
            author=self.author_1,
            role=self.role_1,
            content=self.content_1,
            mention=self.mention_1,
            function=self.function_1,
            timestamp=self.timestamp_1,
            success=self.success_1
        )

        self.po_2 = PersistenceObject(
            uid=self.uid_2,
            author=self.author_2,
            role=self.role_2,
            content=self.content_2,
            mention=self.mention_2,
            function=self.function_2,
            timestamp=self.timestamp_2,
            success=self.success_2
        )

    def test_class_identity(self):
        self.assertIsInstance(self.po_1, PersistenceObject)
        self.assertIsInstance(self.po_2, PersistenceObject)

    def test_class_attribute_uid(self):
        self.assertEqual(self.po_1.uid, self.uid_1)
        self.assertEqual(self.po_2.uid, self.uid_2)

    def test_class_attribute_author(self):
        self.assertEqual(self.po_1.author, self.author_1)
        self.assertEqual(self.po_2.author, self.author_2)

    def test_class_attribute_role(self):
        self.assertEqual(self.po_1.role, self.role_1)
        self.assertEqual(self.po_2.role, self.role_2)

    def test_class_attribute_content(self):
        self.assertEqual(self.po_1.content, self.content_1)
        self.assertEqual(self.po_2.content, self.content_2)

    def test_class_attribute_mention(self):
        self.assertEqual(self.po_1.mention, self.mention_1)
        self.assertEqual(self.po_2.mention, self.mention_2)

    def test_class_attribute_function(self):
        self.assertEqual(self.po_1.function, self.function_1)
        self.assertEqual(self.po_2.function, self.function_2)

    def test_class_attribute_timestamp(self):
        self.assertEqual(self.po_1.timestamp, self.timestamp_1)
        self.assertEqual(self.po_2.timestamp, self.timestamp_2)

    def test_class_attribute_success(self):
        self.assertEqual(self.po_1.success, self.success_1)
        self.assertEqual(self.po_2.success, self.success_2)

    def test_class_method_as_message(self):
        self.po_1_m = self.po_1.as_message()
        self.po_2_m = self.po_2.as_message()

        self.assertIn('role', self.po_1_m.keys())
        self.assertIn('content', self.po_1_m.keys())
        self.assertIn('name', self.po_1_m.keys())

        self.assertEqual(self.po_1_m.get('role'), self.po_1.role)
        self.assertEqual(self.po_1_m.get('content'), self.po_1.content)
        self.assertEqual(self.po_1_m.get('name'), self.po_1.function)

        self.assertEqual(self.po_2_m.get('role'), self.po_2.role)
        self.assertEqual(self.po_2_m.get('content'), self.po_2.content)
        self.assertEqual(self.po_2_m.get('name'), self.po_2.function)


class AlterPersistence(BasePersistence):

    def save(self,
             role: str,
             author: str,
             content: str,
             mentions: List[str] = None,
             function: str = None,
             success: bool = True):
        saved_objs = []
        if mentions is None:
            mentions = ['user']

        for mention in mentions:
            data_obj = PersistenceObject(
                uid=generate_id(16),
                role=role,
                author=author,
                content=content,
                mention=mention,
                function=function,
                timestamp=datetime.now(),
                success=success
            )
            saved_objs.append(data_obj)
        return saved_objs


class TestBasePersistence(unittest.TestCase):

    def setUp(self) -> None:
        self.per_1 = BasePersistence()
        self.per_2 = AlterPersistence()

        self.role_2 = generate_id(8)
        self.author_2 = generate_id(8)
        self.content_2 = generate_id(8)
        self.mention_2 = [self.author_2]
        self.function_2 = generate_id(8)

        self.per_2_saved = self.per_2.save(
            role=self.role_2,
            author=self.author_2,
            content=self.content_2,
            mentions=self.mention_2,
            function=self.function_2
        )

        self.per_2_saved_mock = [
            PersistenceObject(
                uid=generate_id(16),
                role=self.role_2,
                author=self.author_2,
                content=self.content_2,
                mention=self.author_2,
                function=self.function_2,
                timestamp=datetime.now()
            )
        ]

    def test_class_identity(self):
        self.assertIsInstance(self.per_1, BasePersistence)
        self.assertIsInstance(self.per_2, AlterPersistence)

    def test_class_method_save(self):
        self.assertEqual(self.per_1.save(
            role='', author='', content=''
        ), [])

        self.assertEqual(len(self.per_2_saved), len(self.per_2_saved_mock))
        self.assertEqual(self.per_2_saved[0].role, self.per_2_saved_mock[0].role)
        self.assertEqual(self.per_2_saved[0].author, self.per_2_saved_mock[0].author)
        self.assertEqual(self.per_2_saved[0].content, self.per_2_saved_mock[0].content)
        self.assertEqual(self.per_2_saved[0].mention, self.per_2_saved_mock[0].mention)
        self.assertEqual(self.per_2_saved[0].function, self.per_2_saved_mock[0].function)
        self.assertEqual(self.per_2_saved[0].success, self.per_2_saved_mock[0].success)



import json
import os
from typing import AnyStr, Tuple

from jam.interface.base import BaseInterface, JIOutput, JIError
from jam.util.generate import generate_id

import requests
from PIL import Image


class WriteSonicBase(BaseInterface):

    def __init__(self,
                 api_key: str = None,
                 product: str = None,
                 engine: str = None,
                 language: str = 'en'):
        super(WriteSonicBase, self).__init__()
        self._api_key = api_key
        if self._api_key is None:
            self._api_key = os.getenv('WRITESONIC_KEY', '')

        if self._api_key == '':
            raise JIError(f"Unfulfilled credentials for {self.__class__.__name__} in parameters or environment.")

        self.headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "X-API-KEY": self._api_key
        }

        self.product = product
        self.base_url = 'https://api.writesonic.com/v2/business/content/{product}'.format(
            product=self.product
        )
        self.engine = engine
        self.language = language


class PhotoSonicText2Image(WriteSonicBase):

    def __init__(self,
                 api_key: str = None,
                 product: str = 'generate-image',
                 engine: str = 'premium',
                 language: str = 'en',
                 num_copies: int = 1,
                 num_images: int = 1,
                 dimension: Tuple[int, int] = (512, 512),
                 ):
        super().__init__(api_key, product, engine, language)
        self.num_copies = num_copies
        self.num_images = num_images
        self.width, self.height = dimension

    def call(self, x: AnyStr = None) -> JIOutput:
        filenames = []

        self.headers['num_copies'] = str(self.num_copies)
        payload = {
            "num_images": self.num_images,
            "image_width": self.width,
            "image_height": self.height,
            "prompt": x
        }

        response = requests.post(self.base_url, json=payload, headers=self.headers)
        json_response = json.loads(response.text)

        if response.status_code == 200:
            for img_url in json_response.get('images', []):
                filename = f'PhotoSonic_{generate_id(8)}.png'
                img = Image.open(requests.get(img_url, stream=True).raw)
                img.save(filename)
                filenames.append(filename)
            return JIOutput(
                out_type=JIOutput.JIO_TEXT,
                output=' '.join(filenames)
            )
        return JIOutput(
            out_type=JIOutput.JIO_FILE,
            output=json_response.get('msg', ''),
            success=False
        )





from datetime import datetime

from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy import DateTime
from sqlalchemy import Boolean
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class ConversationHistory(Base):
    __tablename__ = "conversation_history"

    uid: Mapped[str] = mapped_column(String(16), primary_key=True)
    author: Mapped[str] = mapped_column(String(32))
    role: Mapped[str] = mapped_column(String(32))
    content: Mapped[str] = mapped_column(Text)
    mention: Mapped[str] = mapped_column(String(32), nullable=True)
    function: Mapped[str] = mapped_column(String(32), nullable=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime, default=datetime.now)
    success: Mapped[bool] = mapped_column(Boolean, default=True)



from typing import Any, AnyStr


class JIError(Exception):

    def __init__(self, message: str):
        self.message = message


class JIOutput(object):
    JIO_TEXT = 'text'
    JIO_IMAGE = 'image'
    JIO_AUDIO = 'audio'
    JIO_VIDEO = 'video'
    JIO_FILE = 'file'
    JIO_OTHER = 'other'

    def __init__(self, out_type: str = JIO_TEXT, output: Any = '', function: AnyStr = None, success: bool = True):
        self.out_type = out_type
        self.output = output
        self.function = function
        self.success = success

    def __repr__(self):
        return f'<JIOutput output:{self.output}, function:{self.function}, success:{self.success}>'


class BaseInterface(object):

    def __init__(self):
        pass

    def call(self) -> JIOutput:
        return JIOutput()



import unittest

from jam.instrument.base import BaseProperty
from jam.instrument.base import BaseParameter
from jam.instrument.base import BaseFunction
from jam.instrument.base import BaseInstrument

from jam.interface.base import BaseInterface


class TestBaseProperty(unittest.TestCase):

    def setUp(self) -> None:
        self.a_property = BaseProperty(name='A', obj_type='int', description='A', enum=['X', 'Y'])
        self.a_property_dict = {
            'A': {
                'type': 'int',
                'description': 'A',
                'enum': ['X', 'Y']
            }
        }
        self.a_property_dict_val = {
            'type': 'int',
            'description': 'A',
            'enum': ['X', 'Y']
        }

        self.b_property = BaseProperty(name='B', description='B')
        self.b_property_dict = {
            'B': {
                'type': 'string',
                'description': 'B'
            }
        }
        self.b_property_dict_val = {
            'type': 'string',
            'description': 'B'
        }

    def test_class_identity(self):
        self.assertIsInstance(self.a_property, BaseProperty)
        self.assertIsInstance(self.b_property, BaseProperty)

    def test_class_equality(self):
        self.assertEqual(self.a_property.as_dict(), self.a_property_dict)
        self.assertEqual(self.a_property.as_dict(value_only=True), self.a_property_dict_val)

        self.assertEqual(self.b_property.as_dict(), self.b_property_dict)
        self.assertEqual(self.b_property.as_dict(value_only=True), self.b_property_dict_val)

    def test_class_inequality(self):
        self.assertNotEqual(self.a_property, self.a_property_dict)
        self.assertNotEqual(self.b_property, self.b_property_dict)

    def tearDown(self) -> None:
        pass


class TestBaseParameter(unittest.TestCase):

    def setUp(self) -> None:
        self.a_property = BaseProperty(name='A', obj_type='int', description='A', enum=['X', 'Y'])
        self.b_property = {'B': {'type': 'string', 'description': 'B'}}

        self.a_parameter = BaseParameter(properties=[self.a_property, self.b_property])
        self.b_parameter = BaseParameter(properties=[self.a_property], required=[self.a_property])
        self.c_parameter = BaseParameter(properties=[self.a_property], required=['A'])
        self.d_parameter = BaseParameter(properties=[self.b_property], required=['B'])

        self.a_parameter_dict = {
            'type': 'object',
            'properties': {self.a_property.name: self.a_property.as_dict(value_only=True),
                           'B': {'type': 'string', 'description': 'B'}},
            'required': []
        }
        self.b_parameter_dict = {
            'type': 'object',
            'properties': self.a_property.as_dict(),
            'required': [self.a_property.name]
        }
        self.c_parameter_dict = {
            'type': 'object',
            'properties': self.a_property.as_dict(),
            'required': ['A']
        }
        self.d_parameter_dict = {
            'type': 'object',
            'properties': self.b_property,
            'required': ['B']
        }

    def test_class_identity(self):
        self.assertIsInstance(self.a_parameter, BaseParameter)
        self.assertIsInstance(self.b_parameter, BaseParameter)
        self.assertIsInstance(self.c_parameter, BaseParameter)
        self.assertIsInstance(self.d_parameter, BaseParameter)

    def test_class_equality(self):
        self.assertEqual(self.a_parameter.as_dict(), self.a_parameter_dict)
        self.assertEqual(self.b_parameter.as_dict(), self.b_parameter_dict)
        self.assertEqual(self.c_parameter.as_dict(), self.c_parameter_dict)
        self.assertEqual(self.d_parameter.as_dict(), self.d_parameter_dict)

    def test_class_inequality(self):
        self.assertNotEqual(self.a_parameter, self.a_parameter_dict)
        self.assertNotEqual(self.b_parameter, self.b_parameter_dict)
        self.assertNotEqual(self.c_parameter, self.c_parameter_dict)
        self.assertNotEqual(self.d_parameter, self.d_parameter_dict)

    def test_class_attributes_property(self):
        self.assertEqual(self.a_parameter.properties, self.a_parameter_dict['properties'])
        self.assertEqual(self.b_parameter.properties, self.b_parameter_dict['properties'])
        self.assertEqual(self.c_parameter.properties, self.c_parameter_dict['properties'])
        self.assertEqual(self.d_parameter.properties, self.d_parameter_dict['properties'])

        self.assertEqual(self.b_parameter.properties, self.c_parameter.properties)
        self.assertNotEqual(self.a_parameter.properties, self.d_parameter.properties)

    def test_class_attributes_required(self):
        self.assertEqual(self.a_parameter.required, self.a_parameter_dict['required'])
        self.assertEqual(self.b_parameter.required, self.b_parameter_dict['required'])
        self.assertEqual(self.c_parameter.required, self.c_parameter_dict['required'])
        self.assertEqual(self.d_parameter.required, self.d_parameter_dict['required'])

        self.assertIn(self.a_property.name, self.b_parameter.required)
        self.assertNotIn(self.a_property.name, self.d_parameter.required)

    def tearDown(self) -> None:
        pass


class TestBaseFunction(unittest.TestCase):

    def setUp(self) -> None:
        self.a_property = BaseProperty(name='A', obj_type='int', description='A', enum=['X', 'Y'])
        self.b_property = {'B': {'type': 'string', 'description': 'B'}}

        self.a_parameter = BaseParameter(properties=[self.a_property, self.b_property])
        self.b_parameter = BaseParameter(properties=[self.a_property], required=[self.a_property])

        self.a_function = BaseFunction(name='X', description='X', parameters=self.a_parameter)
        self.b_function = BaseFunction(name='Y', description='Y', parameters=self.b_parameter)

        self.a_function_dict = {
            'name': 'X',
            'description': 'X',
            'parameters': {
                'type': 'object',
                'properties': {self.a_property.name: self.a_property.as_dict(value_only=True),
                               'B': {'type': 'string', 'description': 'B'}},
                'required': []
            }
        }

        self.b_function_dict = {
            'name': 'Y',
            'description': 'Y',
            'parameters': {
                'type': 'object',
                'properties': self.a_property.as_dict(),
                'required': [self.a_property.name]
            }
        }

    def test_class_identity(self):
        self.assertIsInstance(self.a_function, BaseFunction)
        self.assertIsInstance(self.b_function, BaseFunction)

    def test_class_equality(self):
        self.assertEqual(self.a_function.as_dict(), self.a_function_dict)
        self.assertEqual(self.b_function.as_dict(), self.b_function_dict)

    def test_class_inequality(self):
        self.assertNotEqual(self.a_function, self.a_function_dict)
        self.assertNotEqual(self.b_function, self.b_function_dict)

    def test_class_attributes_parameters(self):
        self.assertEqual(self.a_function.parameters, self.a_function_dict['parameters'])
        self.assertEqual(self.b_function.parameters, self.b_function_dict['parameters'])

        self.assertNotEqual(self.a_function.parameters, self.b_function.parameters)

    def tearDown(self) -> None:
        pass


class TestBaseInstrument(unittest.TestCase):

    def setUp(self) -> None:
        self.a_property = BaseProperty(name='A', obj_type='int', description='A', enum=['X', 'Y'])
        self.b_property = {'B': {'type': 'string', 'description': 'B'}}

        self.a_parameter = BaseParameter(properties=[self.a_property, self.b_property])
        self.b_parameter = BaseParameter(properties=[self.a_property], required=[self.a_property])

        self.a_interface = BaseInterface()
        self.b_interface = BaseInterface()

        self.a_instrument = BaseInstrument(name='X', description='X', parameters=self.a_parameter,
                                           interface=self.a_interface)
        self.b_instrument = BaseInstrument(name='Y', description='Y', parameters=self.b_parameter,
                                           interface=self.b_interface)

        self.a_instrument_dict = {
            'name': 'X',
            'description': 'X',
            'parameters': {
                'type': 'object',
                'properties': {self.a_property.name: self.a_property.as_dict(value_only=True),
                               'B': {'type': 'string', 'description': 'B'}},
                'required': []
            }
        }

        self.b_instrument_dict = {
            'name': 'Y',
            'description': 'Y',
            'parameters': {
                'type': 'object',
                'properties': self.a_property.as_dict(),
                'required': [self.a_property.name]
            }
        }

    def test_class_identity(self):
        self.assertIsInstance(self.a_instrument, BaseInstrument)
        self.assertIsInstance(self.a_instrument, BaseInstrument)

    def test_class_equality(self):
        self.assertEqual(self.a_instrument.as_dict(), self.a_instrument_dict)
        self.assertEqual(self.b_instrument.as_dict(), self.b_instrument_dict)

    def test_class_inequality(self):
        self.assertNotEqual(self.a_instrument, self.a_instrument_dict)
        self.assertNotEqual(self.b_instrument, self.b_instrument_dict)

    def test_class_attributes_parameters(self):
        self.assertEqual(self.a_instrument.parameters, self.a_instrument_dict['parameters'])
        self.assertEqual(self.b_instrument.parameters, self.b_instrument_dict['parameters'])

        self.assertNotEqual(self.a_instrument.parameters, self.b_instrument.parameters)

    def test_class_attributes_interface(self):
        self.assertEqual(self.a_instrument.interface, self.a_interface)
        self.assertEqual(self.b_instrument.interface, self.b_interface)

        self.assertNotEqual(self.a_instrument.interface, self.b_interface)
        self.assertNotEqual(self.b_instrument.interface, self.a_interface)

    def tearDown(self) -> None:
        pass



from typing import List, Dict

from jam.persistence.base import BasePersistence
from jam.persistence.base import PersistenceObject
from jam.persistence.model import Base
from jam.persistence.model import ConversationHistory
from jam.util.generate import generate_id

from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy import or_, and_
from sqlalchemy.orm import Session


class PostgresPersistence(BasePersistence):

    def __init__(self, dbname: str = 'postgresql+psycopg2://localhost:5432/postgres'):
        super(PostgresPersistence, self).__init__()
        self._dbname = dbname
        self.db = create_engine(self._dbname)
        self._metadata = MetaData()

        self.seed()

    def seed(self):
        if not self.db.dialect.has_table(self.db.connect(), 'conversation_history'):
            Base.metadata.create_all(self.db)

    @staticmethod
    def transform(data: ConversationHistory) -> PersistenceObject:
        data_obj = PersistenceObject(
            uid=data.uid,
            role=data.role,
            author=data.author,
            content=data.content,
            mention=data.mention,
            function=data.function,
            timestamp=data.timestamp,
            success=data.success
        )
        return data_obj

    def save(self,
             role: str,
             author: str,
             content: str,
             mentions: List[str] = None,
             function: str = None,
             success: bool = True):
        saved_objs = []
        if mentions is None:
            mentions = ['user']

        with Session(self.db) as session:
            for mention in mentions:
                conv_history = ConversationHistory(
                    uid=generate_id(16),
                    role=role,
                    author=author,
                    content=content,
                    mention=mention,
                    function=function,
                    success=success
                )
                session.add(conv_history)
                saved_objs.append(conv_history)
            session.commit()

            return list(map(self.transform, saved_objs))

    def find(self, key: str, value: List[str] = None, limit: int = 5):
        with Session(self.db) as session:
            key_conditions = [getattr(ConversationHistory, key) == val for val in value]
            key_conditions = or_(*key_conditions)

            mention_conditions = [getattr(ConversationHistory, 'mention') == val for val in value]
            mention_conditions = or_(*mention_conditions)

            filter_conditions = and_(key_conditions, mention_conditions)

            result = session.query(
                ConversationHistory
            ).filter(filter_conditions).order_by(ConversationHistory.timestamp.desc()).limit(limit).all()

            return list(map(self.transform, result[::-1]))

    def all(self):
        with Session(self.db) as session:
            result = session.query(ConversationHistory).all()
            return list(map(self.transform, result))

    def count(self):
        with Session(self.db) as session:
            result = session.query(ConversationHistory).count()
            return result

    def clear(self, key: str = None, value: List[str] = None):
        with Session(self.db) as session:
            if key is None or value is None:
                session.query(ConversationHistory).delete()
                session.commit()
                return

            filter_conditions = [getattr(ConversationHistory, key) == val for val in value]
            filter_conditions = or_(*filter_conditions)

            session.query(ConversationHistory).filter(filter_conditions).delete()
            session.commit()
            return




from typing import Optional, List

from jam.instrument.base import BaseInstrument
from jam.personnel.base import BasePersonnel
from jam.persistence.memory import MemoryPersistence


class BaseJam(object):

    def __init__(self,
                 members: Optional[List[BasePersonnel]],
                 instruments: List[BaseInstrument] = None,
                 persistence: Optional[str] = None):
        self.members = members
        self.instruments = instruments

        self.persistence = persistence
        if self.persistence is None:
            self.persistence = MemoryPersistence()

        self.setup()

    def setup(self):
        for member in self.members:
            member.setup(self.instruments, self.persistence)

    def compose(self, message: str):
        pass



import unittest

from typing import AnyStr

from jam.interface.base import JIOutput
from jam.interface.base import BaseInterface


class TestJIOutput(unittest.TestCase):

    def setUp(self) -> None:
        self.ji_1 = JIOutput()
        self.ji_2 = JIOutput(out_type=JIOutput.JIO_FILE, output='any.file', function='file_maker')
        self.ji_3 = JIOutput(success=False)

    def test_class_identity(self):
        self.assertIsInstance(self.ji_1, JIOutput)
        self.assertIsInstance(self.ji_2, JIOutput)
        self.assertIsInstance(self.ji_3, JIOutput)

    def test_class_attribute_out_type(self):
        self.assertEqual(self.ji_1.out_type, JIOutput.JIO_TEXT)
        self.assertEqual(self.ji_2.out_type, JIOutput.JIO_FILE)
        self.assertEqual(self.ji_3.out_type, JIOutput.JIO_TEXT)

    def test_class_attribute_output(self):
        self.assertEqual(self.ji_1.output, '')
        self.assertEqual(self.ji_2.output, 'any.file')
        self.assertEqual(self.ji_3.output, '')

    def test_class_attribute_function(self):
        self.assertEqual(self.ji_1.function, None)
        self.assertEqual(self.ji_2.function, 'file_maker')
        self.assertEqual(self.ji_3.function, None)

    def test_class_attribute_success(self):
        self.assertEqual(self.ji_1.success, True)
        self.assertEqual(self.ji_2.success, True)
        self.assertEqual(self.ji_3.success, False)

    def tearDown(self) -> None:
        pass


class AlterInterface(BaseInterface):

    def call(self, x: AnyStr = None) -> JIOutput:
        return JIOutput(output=x, success=False)


class TestBaseInterface(unittest.TestCase):

    def setUp(self) -> None:
        self.bi_1 = BaseInterface()
        self.bi_2 = AlterInterface()

    def test_class_identity(self):
        self.assertIsInstance(self.bi_1, BaseInterface)
        self.assertIsInstance(self.bi_2, BaseInterface)

    def test_class_method_call(self):
        self.assertIsInstance(self.bi_1.call(), JIOutput)
        self.assertIsInstance(self.bi_2.call(), JIOutput)

        bi_1_call = self.bi_1.call()
        self.assertEqual(bi_1_call.out_type, JIOutput.JIO_TEXT)
        self.assertEqual(bi_1_call.output, '')
        self.assertEqual(bi_1_call.function, None)
        self.assertEqual(bi_1_call.success, True)

        bi_2_call = self.bi_2.call(x='Hello')
        self.assertEqual(bi_2_call.out_type, JIOutput.JIO_TEXT)
        self.assertEqual(bi_2_call.output, 'Hello')
        self.assertEqual(bi_2_call.function, None)
        self.assertEqual(bi_2_call.success, False)





from typing import List, Dict, Union, Any

from jam.interface.base import BaseInterface


class BaseProperty(object):

    def __init__(self, name: str, obj_type: str = 'string', description: str = None, enum: List[Any] = None):
        self.name = name
        self.type = obj_type
        self.description = description
        self.enum = enum

    def as_dict(self, value_only=False):
        key = ['type', 'description', 'enum']
        val = [self.type, self.description, self.enum]

        if value_only:
            return {k: v for (k, v) in zip(key, val) if v}
        return {
            self.name: {k: v for (k, v) in zip(key, val) if v}
        }


class BaseParameter(object):

    def __init__(self,
                 obj_type: str = 'object',
                 properties: List[Union[BaseProperty, Dict]] = None,
                 required: List[Union[BaseProperty, str]] = None
                 ):
        self.type = obj_type

        self._prop = properties
        self.properties = self._structure_property()

        self._req = required or []
        self.required = self._structure_required()

    def _structure_required(self):
        return [
            x.name if isinstance(x, BaseProperty) else x for x in self._req
        ]

    def _structure_property(self):
        _prop = {}
        for p in self._prop:
            if isinstance(p, BaseProperty):
                _prop[p.name] = p.as_dict(value_only=True)
                continue
            k = [*p][0]  # Get Key
            v = p[k]  # Get Value
            _prop[k] = v
        return _prop

    def as_dict(self):
        return {
            'type': self.type,
            'properties': self.properties,
            'required': self.required
        }


class BaseFunction(object):

    def __init__(self,
                 name: str,
                 description: str,
                 parameters: Union[BaseParameter, Dict]
                 ):
        self.name = name
        self.description = description

        self._param = parameters
        self.parameters = self._structure_parameters()

    def _structure_parameters(self):
        return self._param.as_dict() if isinstance(self._param, BaseParameter) else self._param

    def as_dict(self):
        return {
            'name': self.name,
            'description': self.description,
            'parameters': self.parameters
        }


class BaseInstrument(BaseFunction):

    def __init__(self,
                 name: str,
                 description: str,
                 parameters: Union[BaseParameter, Dict],
                 interface: BaseInterface):
        super().__init__(name, description, parameters)
        self.interface = interface

    def setup(self):
        return

    def activate(self):
        return



from jam import Jam
from jam.personnel import BasicPersonnel
from jam.instrument import PromptPainter
from jam.persistence.sqlite import SQLitePersistence

from rich.console import Console
from rich.prompt import Prompt, Confirm
from rich.text import Text
from rich.panel import Panel


def main():
    console = Console()
    color_wheel = ["red", "green", "blue", "yellow", "cyan", "dark_orange"]

    # Welcome message
    panel_text = """
    🍓 Create a Jam Session with AI.\nhttps://github.com/abhishtagatya/jam
    """
    panel = Panel(
        Text(panel_text, justify='center', style='grey100'),
        title=f"[grey100]Jam v{Jam.__version__}",
        subtitle="[grey100]Let's Jam!",
        style='purple'
    )
    console.print(panel)
    console.print()

    jam_room = Jam(
        members=[
            BasicPersonnel.from_json('example/personnel/homer-simpson.json'),
            BasicPersonnel.from_json('example/personnel/walter-white.json')
        ],
        instruments=[
            PromptPainter()
        ],
        # persistence=SQLitePersistence()
    )

    color_map = {x.uid: color_wheel.pop() for x in jam_room.members}
    for member in jam_room.members:
        tc = color_map[member.uid]
        console.print(f"[{tc}]{member.uid}[/{tc}] has entered the room!")

    # Handle user input
    while True:
        user_input = Prompt.ask()
        if str(user_input).lower() == "q":
            user_confirm = Confirm.ask("Leaving the Room? [y/n]")
            if user_confirm:
                break

        # Add user input to the chat history
        messages = jam_room.compose(message=str(user_input), multi=True)

        # Display chat history
        for message in messages:
            tc = color_map[message.author]
            console.print(f"[{tc}]@{message.author}[/{tc}]: {message.content}")

    console.print("You left the room!", style="blue")


if __name__ == "__main__":
    main()



import os
import json
import pathlib
from typing import List, Dict, Any

from jam.instrument.base import BaseInstrument
from jam.persistence.base import BasePersistence
from jam.persistence.base import PersistenceObject
from jam.interface.openai import OpenAIChat
from jam.util.generate import generate_id


class BasePersonnel(object):
    BASE_TEMPLATE = {
        'uid': '<UNIQUE ID>',
        'display_name': '<DISPLAY NAME>',
        'description': '<DESCRIPTION>',
        'categories': ['<CAT-1>', '<CAT-2>']
    }

    def __init__(self,
                 uid: str,
                 display_name: str,
                 description: str,
                 categories: List[str],
                 instruction: str,
                 restriction: str,
                 extra: Dict,
                 mention_only: bool = False):
        self.uid = uid
        self.display_name = display_name
        self.description = description
        self.categories = categories
        self.instruction = instruction
        self.restriction = restriction
        self.extra = extra

        self.instruments = None
        self.persistence = None
        self.functions = None
        self.function_map = None
        self.engine = None

        self.mention_only = mention_only

        self.prompt = self.generate_prompt()

    def setup(self, instruments: List[BaseInstrument], persistence: BasePersistence):
        self.instruments = instruments
        self.persistence = persistence

        self.functions = self._structure_functions()
        self.function_map = self._structure_function_map()
        self.engine = OpenAIChat(functions=self.functions)

    def _structure_functions(self):
        return [func.as_dict() for func in self.instruments]

    def _structure_function_map(self):
        return {func.name: func.activate for func in self.instruments}

    def _retrieve_history(self):
        a = (self.persistence.find('author', [self.uid, 'user']))
        return list(map(PersistenceObject.as_message, a))

    def _retrieve_context(self):
        return [{'author': 'system', 'role': 'system', 'content': self.prompt}] + self._retrieve_history()

    def _save_message(self,
                      role: str,
                      author: str,
                      content: str,
                      function: str = None,
                      success: bool = True) -> PersistenceObject:
        return self.persistence.save(
            role=role,
            author=author,
            content=content,
            function=function,
            success=success
        )

    def call(self, x: Any, functional: bool = True) -> List[PersistenceObject]:
        po_outputs = []

        ctx_chain = self._retrieve_context()
        response = self.engine.call(messages=ctx_chain, functional=functional)

        if response.output.get('function_call'):
            f_name, f_response = self.function_call(response.output)

            if f_response.success:
                f_chain = [{'role': 'function', 'content': str(f_response.output), 'name': str(f_name)}]
                ctx_chain = self._retrieve_context() + f_chain
                after_response = self.engine.call(messages=ctx_chain, functional=False)
                saved_response = self._save_message(
                    role='assistant',
                    author=self.uid,
                    content=after_response.output['content'],
                    success=after_response.success
                )
                po_outputs += saved_response

                function_response = self._save_message(
                    role='function',
                    author=self.uid,
                    content=f_response.output,
                    function=f_name,
                    success=f_response.success
                )
                po_outputs += function_response
        else:
            saved_response = self._save_message(
                role='assistant',
                author=self.uid,
                content=response.output['content'],
                success=response.success
            )
            po_outputs += saved_response

        return po_outputs

    def function_call(self, x: Any):
        f_name = x['function_call']['name']
        f_call = self.function_map[f_name]
        f_args = json.loads(x['function_call']['arguments'])
        f_response = f_call(**f_args)

        return f_name, f_response

    @classmethod
    def get_template(cls, filepath: str = 'new-personnel.json'):
        filename, file_extension = os.path.splitext(filepath)
        if file_extension == '.json':
            with open(filepath, 'w') as f:
                json.dump(cls.BASE_TEMPLATE, f)
            return True
        return False

    def generate_prompt(self):
        prompt = "INSTRUCTION:\n{instruction}\n\n".format(instruction=self.instruction)
        prompt += "RESTRICTIONS:\n{restriction}\n\n".format(restriction=self.restriction)
        prompt += "CHARACTER:\nName: {display_name}\nDescription: {description}\nCategories: {categories}\n\n".format(
            display_name=self.display_name,
            description=self.description,
            categories=', '.join(self.categories)
        )
        prompt += "EXTRA:\n{extra}\n".format(
            extra='\n'.join([
                '{key}: {value}'.format(
                    key=key.title().replace('_', ' '),
                    value=value
                ) for key, value in self.extra.items()
            ])
        )

        return prompt

    @classmethod
    def from_dict(cls, data: Dict, mention_only: bool = False):
        """
        Load Character from Dict

        :param instruments:
        :param mention_only:
        :param data: Dict
        :return: Instance
        """

        uid = data.get('uid', generate_id(8))
        display_name = data.get('display_name', 'No Name')
        description = data.get('description', '')
        categories = data.get('categories', [])
        instruction = data.get('instruction', '')
        restriction = data.get('restriction', '')
        extra = data.get('extra', {})

        return cls(
            uid=uid,
            display_name=display_name,
            description=description,
            categories=categories,
            instruction=instruction,
            restriction=restriction,
            extra=extra,
            mention_only=mention_only
        )

    @classmethod
    def from_json(cls, filepath: str = None, mention_only: bool = False):
        if os.path.exists(filepath):
            ext_file_df = pathlib.Path(filepath).suffix
            if ext_file_df == '.json' or ext_file_df == '.jsonl':
                with open(filepath, 'r') as config_file:
                    config_json = json.load(config_file)

                return cls.from_dict(data=config_json, mention_only=mention_only)
        raise ValueError(f'File {filepath} is not JSON. Please try another format.')



from multiprocessing.pool import ThreadPool
from typing import List, Optional

from jam.base import BaseJam
from jam.version import version
from jam.instrument.base import BaseInstrument
from jam.persistence.base import BasePersistence
from jam.personnel.base import BasePersonnel
from jam.util.search import search_mention


class Jam(BaseJam):

    __version__ = version

    def __init__(self,
                 members: Optional[List[BasePersonnel]],
                 instruments: List[BaseInstrument] = None,
                 persistence: Optional[BasePersistence] = None):
        super(Jam, self).__init__(members, instruments, persistence)
        self.cache_output = []

    @staticmethod
    def _mention_filter(message: str, members: List[BasePersonnel]) -> List[BasePersonnel]:
        mention_filter = [member for member in members if member.uid in search_mention(message)]
        return mention_filter

    def _assign(self, member: BasePersonnel, message: str):
        call_output = member.call(message)
        self.cache_output += call_output
        return

    def _multi_assign(self, args):
        self._assign(*args)
        return

    def compose(self, message: str, multi=True):
        self.cache_output = []

        prompt_member = self._mention_filter(
            message, self.members
        ) or [member for member in self.members if not member.mention_only]
        self.persistence.save(
            role='user',
            author='user',
            content=message,
            mentions=[member.uid for member in prompt_member]
        )

        if multi:
            tasks = [(member, message) for member in prompt_member]
            with ThreadPool() as pool:
                pool.map(self._multi_assign, tasks)
        else:
            for member in prompt_member:
                self._assign(member, message)
        return self.cache_output

    def history(self):
        result = self.persistence.all()
        return result

    def clear(self, key: str = None,  value: List[str] = None):
        self.persistence.clear(key, value)
        return



from datetime import datetime
from typing import Dict, AnyStr, List


class BasePersistence(object):

    def __init__(self):
        pass

    @staticmethod
    def transform(data: Dict) -> Dict:
        return {}

    def save(self,
             role: str,
             author: str,
             content: str,
             mentions: List[str] = None,
             function: str = None,
             success: bool = True):
        return []

    def find(self, key: str, value: str = None, limit: int = 5):
        return []

    def all(self):
        return []

    def count(self):
        return 0

    def clear(self):
        return


class PersistenceObject:

    def __init__(self,
                 uid: AnyStr,
                 author: str,
                 role: str,
                 content: AnyStr,
                 mention: AnyStr,
                 function: AnyStr,
                 timestamp: datetime = None,
                 success: bool = True):
        self.uid = uid
        self.author = author
        self.role = role
        self.content = content
        self.mention = mention
        self.function = function
        self.timestamp = timestamp
        self.success = success

    def as_message(self):
        return {
            'role': self.role,
            'content': self.content,
            'name': self.function
        }

    def __repr__(self):
        return f'<PersistenceObject (author={self.author}, content={self.content}, mention={self.mention})'



from datetime import datetime
from typing import List, Optional

from jam.persistence.base import BasePersistence
from jam.persistence.base import PersistenceObject
from jam.util.generate import generate_id


class Node:

    def __init__(self, data: PersistenceObject):
        self.data = data
        self.prev = None
        self.next = None


class DoubleLinkedList:

    def __init__(self):
        self.head: Optional[Node] = None
        self.tail: Optional[Node] = None
        self.length = 0

    def append(self, data: PersistenceObject):
        new_node = Node(data)
        if not self.head:
            self.head = new_node
            self.tail = new_node
        else:
            # current = self.head
            # while current.next:
            #     current = current.next
            # current.next = new_node
            # new_node.prev = current
            # self.tail = new_node
            current = self.tail
            current.next = new_node
            new_node.prev = self.tail
            self.tail = new_node
        self.length += 1

    def find(self, key: str, value: List[str], limit: int = 5):
        result = []
        current = self.tail
        if current is None:
            return result

        while current is not None and len(result) < limit:
            if current.data.__dict__[key] in value and current.data.mention in value:
                result.insert(0, current.data)
            current = current.prev
        return result

    def all(self):
        result = []
        current = self.tail
        while current is not None:
            result.append(current.data)
            current = current.prev
        return result

    def clear(self, key: str = None, value: List[str] = None):
        if key is None or value is None:
            self.head = self.tail = None
            self.length = 0
        else:
            current = self.tail
            while current:
                if current.data.__dict__[key] in value:
                    if current.prev:
                        current.prev.next = current.next
                    else:
                        self.head = current.next

                    if current.next:
                        current.next.prev = current.prev
                    else:
                        self.tail = current.prev
                    self.length -= 1

                current = current.prev


class MemoryPersistence(BasePersistence):

    def __init__(self):
        super(MemoryPersistence, self).__init__()
        self.db: DoubleLinkedList = DoubleLinkedList()

    def save(self,
             role: str,
             author: str,
             content: str,
             mentions: List[str] = None,
             function: str = None,
             success: bool = True):
        saved_objs = []
        if mentions is None:
            mentions = ['user']

        for mention in mentions:
            data_obj = PersistenceObject(
                uid=generate_id(16),
                role=role,
                author=author,
                content=content,
                mention=mention,
                function=function,
                timestamp=datetime.now(),
                success=success
            )
            self.db.append(data_obj)
            saved_objs.append(data_obj)
        return saved_objs

    def find(self, key: str, value: List[str] = None, limit: int = 5):
        result = self.db.find(key, value, limit)
        return result

    def all(self):
        result = self.db.all()
        return result

    def count(self):
        return self.db.length

    def clear(self, key: str = None, value: List[str] = None):
        self.db.clear(key=key, value=value)
        return




import string
import random


def generate_id(length: int = 8) -> str:
    source = string.ascii_letters + string.digits
    result_str = ''.join((random.choice(source) for i in range(length)))
    return result_str
